/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gpdef.h>
#include <gpstdlib.h>
#include <gpgraphic.h>
#include <gpstdio.h>
#include <gpfont.h>

#include "types.h"
#include "xroar.h"
#include "pia.h"
#include "keyboard.h"
#include "snapshot.h"
#include "logging.h"
#include "video.h"
#include "joystick.h"

#define KEY_PRESSED(s) keyboard_column[keymap[s][0]] &= ~(1<<keymap[s][1]); \
	keyboard_row[keymap[s][1]] &= ~(1<<keymap[s][0])
#define KEY_RELEASED(s) keyboard_column[keymap[s][0]] |= 1<<keymap[s][1]; \
	keyboard_row[keymap[s][1]] |= 1<<keymap[s][0]
#define KEY_UPDATE(t,s) if (t) { KEY_PRESSED(s); } else { KEY_RELEASED(s); }

#define KEYBOARD_OFFSET (64*240)

extern PIXEL kbd_bin[2][5060];
extern PIXEL cmode_bin[4][1024];

/* This struct and array define where keys are displayed, and what keysym they
 * correspond to. */

typedef struct {
	Uint32 xoffset, width, sym;
} keydata;

static keydata keys[5][14] = {
	{ {0,0,0}, {0,0,0}, {23,7,32}, {31,7,32}, {39,7,32}, {47,7,32}, {55,7,32}, {63,7,32}, {71,7,32}, {79,7,32}, {255,0,0}, {255,0,0}, {255,0,0}, {255,0,0} },
	{ {3,11,0}, {15,7,122}, {23,7,120}, {31,7,99}, {39,7,118}, {47,7,98}, {55,7,110}, {63,7,109}, {71,7,44}, {79,7,46}, {87,7,47}, {95,11,0}, {255,0,0}, {255,0,0} },
	{ {3,7,10}, {11,7,97}, {19,7,115}, {27,7,100}, {35,7,102}, {43,7,103}, {51,7,104}, {59,7,106}, {67,7,107}, {75,7,108}, {83,7,59}, {91,15,13}, {107,7,12}, {255,0,0} },
	{ {1,7,11}, {9,7,113}, {17,7,119}, {25,7,101}, {33,7,114}, {41,7,116}, {49,7,121}, {57,7,117}, {65,7,105}, {73,7,111}, {81,7,112}, {89,7,91}, {97,7,8}, {105,7,9} },
	{ {0,0,0}, {5,7,49}, {13,7,50}, {21,7,51}, {29,7,52}, {37,7,53}, {45,7,54}, {53,7,55}, {61,7,56}, {69,7,57}, {77,7,48}, {85,7,45}, {93,7,61}, {101,7,27} }
};

static Uint32 keyx = 1, keyy = 4;
static keydata *current;
static Uint32 oldkey = ~0;
static Uint32 keyboard_mode = 0;
static Uint32 displayed_keyboard = 0;

static void highlight_key(void);

void keyboard_init_arch(void) {
	video_blit_bothbufs(KEYBOARD_OFFSET, (PIXEL *)kbd_bin[0], 115, 44);
	keyboard_mode = 0;
	video_blit_bothbufs(1928, cmode_bin[keyboard_mode], 32, 32);
	displayed_keyboard = 0;
	keyx = 1; keyy = 4;
	current = &keys[keyy][keyx];
	highlight_key();
}

static void highlight_key(void) {
	int i;
	PIXEL *p = VIDEO_CURRENTBASE + KEYBOARD_OFFSET + current->xoffset*240
		+ 4 + keyy * 8;
	for (i = 0; i < current->width; i++) {
		*(p++) ^= ~0; *(p++) ^= ~0; *(p++) ^= ~0; *(p++) ^= ~0;
		*(p++) ^= ~0; *(p++) ^= ~0; *(p++) ^= ~0;
		p += 233;
	}
}

void keyboard_poll(void) {
	Uint32 newkeyx = keyx, newkeyy = keyy;
	Uint32 key = GpKeyGet();
	Uint32 newkey = key & oldkey;
	oldkey = ~key;
	if (newkey & GPC_VK_FL) {
		keyboard_mode++;
		if (keyboard_mode > 3)
			keyboard_mode = 0;
		video_blit_bothbufs(1928, (PIXEL *)cmode_bin[keyboard_mode], 32, 32);
	}
	if ((key & (GPC_VK_FR|GPC_VK_FL)) == (GPC_VK_FR|GPC_VK_FL))
		snapshot_state = 3;  /* reset machine */
	if (!snapshot_state && newkey & GPC_VK_START)
		snapshot_state = 2;
	switch (keyboard_mode) {
		case 3:
			joystick_rightx = joystick_righty = 255;
			joystick_leftx = joystick_lefty = 129;
			if (key & GPC_VK_LEFT) joystick_leftx = 0;
			if (key & GPC_VK_RIGHT) joystick_leftx = 255;
			if (key & GPC_VK_UP) joystick_lefty = 0;
			if (key & GPC_VK_DOWN) joystick_lefty = 255;
			if (key & GPC_VK_FB)
				PIA_TIE_LOW(PIA_0A,2);
			else
				PIA_UNTIE_LOW(PIA_0A,2);
			KEY_UPDATE(key & GPC_VK_FR, 13);
			KEY_UPDATE(key & GPC_VK_FA, 32);
			KEY_UPDATE(key & GPC_VK_SELECT, 112);
			/*
			keyboard_matrix[0][6] = key & GPC_VK_FR;
			keyboard_matrix[7][5] = key & GPC_VK_FA;
			keyboard_matrix[0][4] = key & GPC_VK_SELECT;
			*/
			break;
		case 2:
			joystick_leftx = joystick_lefty = 255;
			joystick_rightx = joystick_righty = 129;
			if (key & GPC_VK_LEFT) joystick_rightx = 0;
			if (key & GPC_VK_RIGHT) joystick_rightx = 255;
			if (key & GPC_VK_UP) joystick_righty = 0;
			if (key & GPC_VK_DOWN) joystick_righty = 255;
			if (key & GPC_VK_FB)
				PIA_TIE_LOW(PIA_0A,1);
			else
				PIA_UNTIE_LOW(PIA_0A,1);
			KEY_UPDATE(key & GPC_VK_FR, 13);
			KEY_UPDATE(key & GPC_VK_FA, 32);
			KEY_UPDATE(key & GPC_VK_SELECT, 112);
			/*
			keyboard_matrix[0][6] = key & GPC_VK_FR;
			keyboard_matrix[7][5] = key & GPC_VK_FA;
			keyboard_matrix[0][4] = key & GPC_VK_SELECT;
			*/
			break;
		case 1:
			KEY_UPDATE(key & GPC_VK_FR, 13);
			KEY_UPDATE(key & GPC_VK_FB, 0);
			KEY_UPDATE(key & GPC_VK_UP, 11);
			KEY_UPDATE(key & GPC_VK_DOWN, 10);
			KEY_UPDATE(key & GPC_VK_LEFT, 8);
			KEY_UPDATE(key & GPC_VK_RIGHT, 9);
			KEY_UPDATE(key & GPC_VK_FA, 32);
			KEY_UPDATE(key & GPC_VK_SELECT, 100);
			/*
			keyboard_matrix[0][6] = key & GPC_VK_FR;
			keyboard_matrix[7][6] = key & GPC_VK_FB;
			keyboard_matrix[3][5] = key & GPC_VK_UP;
			keyboard_matrix[4][5] = key & GPC_VK_DOWN;
			keyboard_matrix[5][5] = key & GPC_VK_LEFT;
			keyboard_matrix[6][5] = key & GPC_VK_RIGHT;
			keyboard_matrix[7][5] = key & GPC_VK_FA;
			keyboard_matrix[4][2] = key & GPC_VK_SELECT;
			*/
			break;
		default:
		case 0:
			KEY_UPDATE(key & GPC_VK_FR, 0);
			//keyboard_matrix[7][6] = key & GPC_VK_FR;
			if (((key&GPC_VK_FR)!=0)^displayed_keyboard) {
				displayed_keyboard ^= 1;
				video_blit_bothbufs(KEYBOARD_OFFSET, (PIXEL *)kbd_bin[displayed_keyboard], 115, 44);
				highlight_key();
			}
			if (key & GPC_VK_UP && keyy < 4)
				newkeyy = keyy + 1;
			if (key & GPC_VK_DOWN && keyy > 0)
				newkeyy = keyy - 1;
			if (key & GPC_VK_LEFT && keyx > 0)
				newkeyx = keyx - 1;
			if (key & GPC_VK_RIGHT && keyx < 13)
				newkeyx = keyx + 1;
			while (keys[newkeyy][newkeyx].xoffset == 0)
				newkeyx++;
			while (keys[newkeyy][newkeyx].xoffset == 255)
				newkeyx--;
			if (newkeyx != keyx || newkeyy != keyy) {
				Uint32 ticks = GpTickCountGet();
				highlight_key();
				KEY_RELEASED(current->sym);
				//keyboard_matrix[current->i][current->j] = 0;
				keyx = newkeyx; keyy = newkeyy;
				current = &keys[keyy][keyx];
				highlight_key();
				while (GpTickCountGet()-ticks<400);
			}
			KEY_UPDATE(key & GPC_VK_FB, current->sym);
			//keyboard_matrix[current->i][current->j] = key & GPC_VK_FB;
			break;
	}
}
